/*
 * Memory sharing abstractions.
 *
 * Copyright (C) 2018, 2019, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/c/mem_alloc.h>
#include <l4/re/c/rm.h>
#include <l4/sys/err.h>
#include <l4/sys/types.h>

#include "cap_alloc.h"
#include "mem_ipc.h"
#include "semaphore.h"



/* Semaphore used if threading has been initialised by ipc_mem_init. */

l4_cap_idx_t ipc_mem_semaphore = L4_INVALID_CAP;



/* Generic memory allocation. */

long ipc_allocate(unsigned long size, void **addr, l4re_ds_t *ds)
{
  /* Allocate a capability for the dataspace. */

  *ds = ipc_cap_alloc();

  if (l4_is_invalid_cap(*ds))
    return -L4_ENOENT;

  /* Allocate and attach the memory for the dataspace. */

  if (ipc_new_dataspace(*ds, size, 0, 0))
  {
    ipc_cap_free_um(*ds);
    return -L4_ENOMEM;
  }

  if (ipc_attach_dataspace(*ds, size, addr))
  {
    ipc_cap_free_um(*ds);
    return -L4_ENOMEM;
  }

  return L4_EOK;
}

/* Create a dataspace. Equivalent to l4re_ma_alloc_align. */

long ipc_new_dataspace(l4_cap_idx_t cap, l4_mword_t size, l4_umword_t flags, l4_umword_t align)
{
  return l4re_ma_alloc_align(size, cap, flags, align);
}

/* Work around API change. */

#ifndef L4RE_RM_SEARCH_ADDR
#define L4RE_RM_SEARCH_ADDR L4RE_RM_F_SEARCH_ADDR | L4RE_RM_F_RW
#endif

/* Attach a dataspace region. Similar to l4re_rm_attach. */

long ipc_attach_dataspace(l4re_ds_t ds, unsigned long size, void **addr)
{
  long err;

  err = ipc_semaphore_down(ipc_mem_semaphore);
  if (err)
    return err;

  err = l4re_rm_attach(addr, size, L4RE_RM_SEARCH_ADDR, ds, 0, L4_PAGESHIFT);

  ipc_semaphore_up(ipc_mem_semaphore);

  return err;
}

/* Detach a dataspace region. Equivalent to l4re_rm_detach_unmap. */

long ipc_detach_dataspace(void *addr)
{
  long err;

  err = ipc_semaphore_down(ipc_mem_semaphore);
  if (err)
    return err;

  err = l4re_rm_detach_unmap((l4_addr_t) addr, L4RE_THIS_TASK_CAP);

  ipc_semaphore_up(ipc_mem_semaphore);

  return err;
}

/* Obtain a dataspace size. Equivalent to l4re_ds_size. */

long ipc_dataspace_size(l4_cap_idx_t cap, unsigned long *size)
{
  *size = l4re_ds_size(cap);
  return L4_EOK;
}

// vim: tabstop=2 expandtab shiftwidth=2
